from . import base
from typing import *

from neobuilder.neobuilder.render import TemplateRenderer
from ..symbols import ProtoMethod


class ModuleBuilder(base.AbstractModuleBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_render_filename(self) -> str:
        return f'{self.module.get_file_full_name()[:-7]}_api.py'

    def get_import_lines(self) -> List[str]:
        import_set = {
            f'from {self.module.get_package()} import {self.module.get_module_name()[:-3]}dc as dc'
        }
        for s in self.module.service_map.values():

            if self.module.is_grpc_file():
                import_set.add(f'from {self.module.get_package()}.{self.module.get_module_name()[:-3]}grpc_receiver import {s.service_descriptor.name}GrpcServicer')

            for m in s.method_map.values():
                # From 3.2
                if m.input.self_import.get_package() != 'google.protobuf':
                    import_set.add(m.input.self_import.get_render_import(self.module.descriptor))

                if m.output.self_import.get_package() != 'google.protobuf':
                    import_set.add(m.output.self_import.get_render_import(self.module.descriptor))

                for f in m.input.field_map.values():
                    if f.msg_import:
                        if self.module.descriptor != f.msg_import.descriptor.file:
                            import_set.add(f.msg_import.get_render_import(self.module.descriptor))

                for f in m.output.field_map.values():
                    if f.msg_import:
                        if self.module.descriptor != f.msg_import.descriptor.file:
                            import_set.add(f.msg_import.get_render_import(self.module.descriptor))

        return sorted(import_set)

    def get_template_context(self) -> Dict:
        d: Dict = super().get_template_context()
        d.update({
            'package': self.module.get_package(),
            'import_lines': self.get_import_lines(),
            'services': self.render_services(),
            'all_clause': [f'{svc.service_descriptor.name}Interface' for svc in self.module.service_map.values()]
        })
        return d

    def render(self) -> str:
        return TemplateRenderer().render_file('interface', **self.get_template_context())


class ServiceBuilder(base.AbstractServiceBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_method_context(self, method: ProtoMethod) -> Dict[str, Any]:
        d = super().get_method_context(method)
        d.update({
            'method_under_name': method.under_name,
            'input_params': method.get_input_params(True),
            'return_type_hint': method.get_return_type_hint(True),
        })
        return d

    def render(self) -> str:
        return TemplateRenderer().render_file('parts/_interface_service', **self.get_template_context())


class MethodBuilder(base.AbstractMethodBuilder):
    def get_template_context(self) -> Dict:
        d = super().get_template_context()
        d.update({
            'method_under_name': self.method.under_name,
            'input_params': self.method.get_input_params(True),
            'return_type_hint': self.method.get_return_type_hint(True),
        })
        return d

    def render(self) -> str:
        return TemplateRenderer().render_file('parts/_interface_method', **self.get_template_context())



__all__ = [
    'TemplateRenderer',
]
from ccptools.structs import Singleton
import jinja2
import os
import sys

import logging
log = logging.getLogger(__name__)

_PY_3_9_PLUS = sys.version_info >= (3, 9)


def _get_resource_path():
    if _PY_3_9_PLUS:
        from importlib import resources
        return str(resources.files('neobuilder').joinpath('data'))
    else:
        import pkg_resources  # noqa
        return str(pkg_resources.resource_filename('neobuilder', 'data'))  # noqa


class TemplateRenderer(metaclass=Singleton):
    def __init__(self):
        self.template_dir = os.path.join(_get_resource_path(), 'templates')

        self._env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                searchpath=self.template_dir
            ),
            lstrip_blocks=True,
            trim_blocks=True,
            undefined=jinja2.ChainableUndefined
        )

    def render_file(self, template_name: str, **kwargs) -> str:
        try:
            # log.debug(f'Rendering {template_name} with: {kwargs!r}')

            template = self._env.get_template(f'{template_name}.jinja2')
            return template.render(**kwargs)
        except Exception as ex:
            log.exception(f'Error in template: {template_name}: {ex!r}')
            raise ex


def main():
    t = TemplateRenderer()
    print(t.render_file('root_init',
                        version='(3, 1, 0, 0)',
                        protoplasm_version='(4, 0, 0, 0)'))


if __name__ == '__main__':
    main()



from __future__ import annotations
__all__ = [
    'ProtoClass',
    'ProtoField',
    'ProtoEnumValue',
    'ProtoEnum',
]

import typing

from neobuilder.descwrap import *

from google.protobuf import json_format
from google.protobuf import descriptor as pro_desc

from protoplasm.casting import castutils
from neobuilder.generators.symbols import modules
from neobuilder.generators.symbols import common


class ProtoClass(object):
    def __init__(self, descriptor: json_format.descriptor.Descriptor, module: modules.ProtoModule, parent_class: typing.Optional[object] = None, force_import=False):
        self.descriptor = descriptor
        self.module = module
        self.indent = module.indent
        self.enum_map: typing.Mapping[str, ProtoEnum] = {}
        self.field_map: typing.Mapping[str, ProtoField] = {}
        self.nested_message_map: typing.Mapping[str, ProtoClass] = {}
        self.parent_class: typing.Optional[ProtoClass] = parent_class
        self.force_import = force_import
        self.self_import = common.ProtoImport(self.descriptor)
        self._load_class()

    def _load_class(self):
        self.enum_map = {e.name: ProtoEnum(e, self.module, self) for e in self.descriptor.enum_types}
        self.field_map = {f.name: ProtoField(f, self, self.force_import) for f in self.descriptor.fields}
        self.nested_message_map = {}
        for n in self.descriptor.nested_types:
            if not (n.has_options and n.GetOptions().map_entry):
                self.nested_message_map[n.name] = ProtoClass(n, self.module, self)

    def get_class_lineage(self) -> str:
        path = self.get_class_name()
        if self.parent_class:
            path = f'{self.parent_class.get_class_lineage()}.{path}'
        return path

    def get_class_import_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => 'DeltaMessage'
        """
        return self.get_class_name()

    def get_class_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => 'DeltaMessage'
        """
        return self.descriptor.name

    def get_under_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => 'delta_message'
        """
        return castutils.humps_to_under(self.get_class_name())

    def get_proto_cls_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => 'DeltaMessage'
        """
        if self.parent_class:
            return f'{self.parent_class.get_proto_cls_name()}.{self.get_class_name()}'
        return self.get_class_name()

    def get_file_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => FileDescriptor('sandbox/test/delta.proto')
        """
        return self.descriptor.file

    def get_proto_module_name(self):
        """
        :return: 'sandbox.test.delta_dc.DeltaMessage' => 'service_with_imported_io_pb2'
        """
        return self.module.get_module_name()

    def get_imports(self):
        import_set = set()
        for f in self.field_map.values():
            if f.msg_import:
                import_set.add(f.msg_import.get_render_import())
        for c in self.nested_message_map.values():
            import_set.update(c.get_imports())
        return import_set

    def output_google_type_proxy(self):
        return {
            'Empty': 'NoReturn'
        }.get(
            self.get_class_name()
        )

    def input_google_type_proxy(self):
        return {
            'Empty': 'None'
        }.get(
            self.get_class_name()
        )

    def protobuf_google_type_proxy(self):
        return f'{self.get_class_name().lower()}_pb2.{self.get_class_name()}'

    def render_dc_type_hint(self, relative_to: modules.ProtoModule = None) -> str:
        relative_to = relative_to or self.module
        if self.descriptor.file.package == 'google.protobuf':
            return self.output_google_type_proxy()
        if not relative_to or self.module.descriptor != self.self_import.descriptor.file:
            return_type = f'{self.self_import.get_import_name()}.{self.get_class_name()}'
            return return_type
        else:
            return f'dc.{self.get_class_name()}'

    def render_pb2_type_hint(self, relative_to: modules.ProtoModule = None) -> str:
        if self.descriptor.file.package == 'google.protobuf':
            return self.protobuf_google_type_proxy()
        else:
            relative_to = relative_to or self.module
            if not relative_to or self.module.descriptor != self.self_import.descriptor.file:
                return f'{self.self_import.get_import_name()}.pb2.{self.get_class_name()}'
            else:
                return f'dc.pb2.{self.get_class_name()}'

    def render_py(self, indent_level=0):
        buf = [
            f'{self.indent*indent_level}@dataclasses.dataclass\n',
            f'{self.indent*indent_level}class {self.get_class_name()}(plasm.DataclassBase):\n',
            f'{self.indent*(indent_level + 1)}__proto_cls__ = pb2.{self.get_proto_cls_name()}\n'
        ]
        if self.enum_map:
            for en in self.enum_map.values():
                buf.append(f'\n{en.render_py(indent_level + 1)}')
            buf.append('\n')

        if self.nested_message_map:
            for nm in self.nested_message_map.values():
                buf.append(f'\n{nm.render_py(indent_level + 1)}')
            buf.append('\n')

        for pf in self.field_map.values():
            buf.append(f'{pf.render_py(indent_level + 1)}\n')

        return ''.join(buf)

    def render_api_args(self, api_file: typing.Optional[pro_desc.FileDescriptor] = None):
        if self.field_map:
            return ', '.join([pf.render_arg_py(api_file or self.descriptor.file) for pf in self.field_map.values()])
        return ''

    def render_api_request_arg(self, method_desc: MethodDescriptorWrapper):
        if self.descriptor.file.package == 'google.protobuf':
            type_hint = self.input_google_type_proxy()
        else:
            type_hint = f'{self.self_import.get_import_name(method_desc.file_descriptor)}.{self.descriptor.name}'

        if method_desc.io_type.is_input_stream():
            return f', request_iterator: Iterable[{type_hint}]'
        return f', {self.get_under_name()}: {type_hint}'

    def render_forwarded_args(self):
        if self.field_map:
            return ', '.join([pf.py_name() for pf in self.field_map.values()])
        return ''

    def _render_forwarded_args__BAD(self):
        return self.get_under_name()

    def render_api_return_args(self, api_file: typing.Optional[pro_desc.FileDescriptor] = None):
        if self.field_map:
            type_hint_list = [pf.get_type_hint(api_file or self.descriptor.file) for pf in self.field_map.values()]
            if len(type_hint_list) == 1:
                return type_hint_list[0]
            else:
                return f"Tuple[{', '.join(type_hint_list)}]"
        return 'NoReturn'

    def render_api_return_dataclass(self, api_file: typing.Optional[pro_desc.FileDescriptor] = None):
        if self.descriptor.file.package == 'google.protobuf':
            return self.output_google_type_proxy()
        return f'{self.self_import.get_import_name(api_file)}.{self.get_class_name()}'


class ProtoField(object):
    def __init__(self,
                 field_descriptor: json_format.descriptor.FieldDescriptor,
                 parent_class: ProtoClass,
                 force_import: bool = False):
        """
        :type field_descriptor: google.protobuf.descriptor.FieldDescriptor
        """
        self.field_descriptor = FieldDescriptorWrapper(field_descriptor)
        self.parent_class = parent_class
        self.indent = parent_class.indent
        self.force_import = force_import

        self.msg_import: typing.Optional[common.ProtoImport] = None

        if self.is_message() and not self.is_any():
            if self.force_import or self.field_descriptor.value_msg.file != self.parent_class.module.descriptor:
                self.msg_import = common.ProtoImport(self.field_descriptor.value_msg)

        if self.is_enum():
            if self.force_import or self.field_descriptor.value_field.enum_type.file != self.parent_class.module.descriptor:
                self.msg_import = common.ProtoImport(self.field_descriptor.value_field.enum_type)

    def is_proto_repeated(self) -> bool:
        return self.field_descriptor.is_proto_repeated

    def py_name(self):
        if self.is_keyword():
            return f'{self.field_descriptor.name}_'
        return self.field_descriptor.name

    def render_py(self, indent_level=0):
        return f'{self.indent*indent_level}{self.py_name()}: {self.get_type_hint()} = dataclasses.field({self.get_field_options()})'

    def render_arg_py(self, used_in_file: json_format.descriptor.FileDescriptor = None):
        return f'{self.py_name()}: {self.get_type_hint(used_in_file)} = None'

    def get_field_options(self):
        if self.is_map():
            return f"default_factory=dict, metadata={{{self.get_metadata()}}}"
        elif self.is_list():
            return f"default_factory=list, metadata={{{self.get_metadata()}}}"
        elif self.is_enum():
            # This breaks in the case of an internal class using internal enum
            # (e.g. eve.starbase.starbase_pb2.Settings)...
            #
            # enum_guy = self.get_py_type_name(self.field_descriptor.real_descriptor.file)
            # if self.field_descriptor.real_descriptor.file == self.field_descriptor.real_descriptor.enum_type.file:
            #     if '.' in enum_guy:  # Nested... break it!
            #         enum_guy = enum_guy[enum_guy.rindex('.')+1:]
            # return f"default={enum_guy}(0), metadata={{{self.get_metadata()}}}"
            #
            # ...so we'll just use 0 for now!
            return f"default=0, metadata={{{self.get_metadata()}}}"
        return f"default={self.field_descriptor.py_type.default_value}, metadata={{{self.get_metadata()}}}"

    def get_py_type(self) -> common.PyType:
        return self.field_descriptor.py_type

    def get_full_path(self, containing_type: pro_desc.Descriptor):
        def get_containing_type(value_msg: pro_desc.Descriptor):
            path = value_msg.name
            if value_msg.containing_type:
                path = f'{get_containing_type(value_msg.containing_type)}.{path}'
            return path
        return get_containing_type(containing_type)

    def get_py_module_name(self, used_in_file: json_format.descriptor.FileDescriptor = None) -> str:
        if self.msg_import:
            return f'{self.msg_import.get_import_name(used_in_file)}.'
        if self.is_enum() and self.field_descriptor.value_field.enum_type.containing_type:
            return f'{self.get_full_path(self.field_descriptor.value_field.enum_type.containing_type)}.'
        return ''

    def get_py_type_name(self, used_in_file: json_format.descriptor.FileDescriptor = None) -> str:
        if self.is_timestamp():
            return 'datetime.datetime'
        if self.is_duration():
            return 'datetime.timedelta'
        if self.is_any():
            return 'plasm.DataclassBase'
        if self.is_message():
            return f'{self.get_py_module_name(used_in_file)}{self.get_message_container_type()}'
        elif self.is_enum():
            return f'{self.get_py_module_name(used_in_file)}{self.field_descriptor.value_field.enum_type.name}'
        else:
            return self.field_descriptor.py_type.name

    def get_type_hint(self, used_in_file: json_format.descriptor.FileDescriptor = None) -> str:
        if self.is_map():
            return f'Dict[{self.field_descriptor.key_py_type.name}, {self.get_py_type_name(used_in_file)}]'
        elif self.is_list():
            return f'List[{self.get_py_type_name(used_in_file)}]'
        elif self.is_empty():
            return f'None'
        return self.get_py_type_name(used_in_file)

    def get_message_container_type(self) -> str:
        def get_containing_type(value_msg: pro_desc.Descriptor):
            path = value_msg.name
            if value_msg.containing_type:
                path = f'{get_containing_type(value_msg.containing_type)}.{path}'
            return path
        return get_containing_type(self.field_descriptor.value_msg)

    def get_metadata(self):
        buf = [f"'dictator': dictators.{self.get_dictator_name()}"]
        if self.is_map():
            buf.append("'is_map': True")
        elif self.is_list():
            buf.append("'is_list': True")

        if self.is_message():
            buf.append("'is_obj': True")
        elif self.is_enum():
            buf.append("'is_enum': True")

        if self.is_keyword():
            buf.append(f"'pb_name': '{self.field_descriptor.name}'")

        return ', '.join(buf)

    def get_dictator_name(self):
        if self.is_empty():
            return 'EmptyDictator'
        if self.is_any():
            return 'AnyDictator'
        if self.is_timestamp():
            return 'TimestampDictator'
        if self.is_duration():
            return 'DurationDictator'
        if self.is_bytes():
            return 'ByteDictator'
        if self.is_enum():
            return 'EnumDictator'
        if self.is_long():
            return 'LongDictator'
        return 'BaseDictator'

    def is_long(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_LONG)

    def is_timestamp(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_TIMESTAMP)

    def is_duration(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_DURATION)

    def is_any(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_ANY)

    def is_empty(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_EMPTY)

    def is_bytes(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_BYTES)

    def is_single(self):
        return bool(self.field_descriptor.kind & FieldKind.KIND_SINGLE)

    def is_map(self):
        return bool(self.field_descriptor.kind & FieldKind.KIND_MAP)

    def is_list(self):
        return bool(self.field_descriptor.kind & FieldKind.KIND_LIST)

    def is_scalar(self):
        return bool(self.field_descriptor.kind & FieldKind.DATA_SCALAR)

    def is_message(self):
        return bool(self.field_descriptor.kind & FieldKind.DATA_MESSAGE)

    def is_enum(self):
        return bool(self.field_descriptor.kind & FieldKind.DATA_ENUM)

    def is_keyword(self):
        return bool(self.field_descriptor.kind & FieldKind.SPECIAL_KEYWORD)


class ProtoEnumValue(object):
    def __init__(self, descriptor: json_format.descriptor.EnumValueDescriptor):
        self.descriptor = descriptor

    def get_name(self):
        return self.descriptor.name

    def get_value(self):
        return self.descriptor.number

    def render_py(self):
        return f'{self.get_name()} = {self.get_value()}'


class ProtoEnum(object):
    def __init__(self, descriptor: json_format.descriptor.EnumDescriptor, module: modules.ProtoModule, parent_class: typing.Optional[ProtoClass] = None):
        self.descriptor = descriptor
        self.module = module
        self.parent_class = parent_class
        self.indent = module.indent

    def get_name(self):
        return self.descriptor.name

    def get_class_import_name(self):
        buf = []
        if self.parent_class:
            buf.append(self.parent_class.get_class_lineage())
        buf.append(self.get_name())
        return '.'.join(buf)

    def render_py_consts(self, indent_level=0):
        buf = []
        for v in reversed(self.descriptor.values):  # Reversing order to ensure default Alias render correctly
            pv = ProtoEnumValue(v)
            buf.append(f'{self.indent*indent_level}{pv.get_name()} = {self.get_name()}({pv.get_value()})\n')
        return ''.join(buf)

    def render_py_class(self, indent_level=0):
        buf = [
            f'{self.indent*indent_level}class {self.get_name()}(enum.IntEnum):\n',
            f'{self.indent*(indent_level + 1)}__proto_cls__ = pb2.{self.get_class_import_name()}\n'
        ]
        for v in reversed(self.descriptor.values):  # Reversing order to ensure default Alias render correctly
            pv = ProtoEnumValue(v)
            buf.append(f'{self.indent*(indent_level + 1)}{pv.render_py()}\n')
        return ''.join(buf)

    def render_py(self, indent_level=0):
        buf = [
            self.render_py_class(indent_level),
            '\n' if self.parent_class else '\n\n',
            self.render_py_consts(indent_level)
        ]
        return ''.join(buf)



from . import base
import grpc
from neobuilder.structs import *
from neobuilder.neobuilder.render import TemplateRenderer


class ModuleBuilder(base.AbstractModuleBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_render_filename(self) -> str:
        return f'{self.module.get_file_full_name()[:-7]}_grpc_receiver.py'

    def render_imports(self) -> str:
        import_set = {
            f'from {self.module.get_package()} import {self.module.get_module_name()} as pb2',
            f'from {self.module.get_package()} import {self.module.get_module_name()}_grpc as pb2_grpc',
            f'from {self.module.get_package()} import {self.module.get_module_name()[:-3]}dc as dc',
        }

        for s in self.module.service_map.values():
            for m in s.method_map.values():
                if m.input:
                    if m.input.self_import:
                        if self.module.descriptor != m.input.self_import.descriptor.file:
                            if m.input.self_import.get_package() != 'google.protobuf':
                                import_set.add(m.input.self_import.get_render_import())
                            else:
                                import_set.add(f'from google.protobuf import {m.input.self_import.get_class_name().lower()}_pb2')
                if m.output:
                    if m.output.self_import:
                        if self.module.descriptor != m.output.self_import.descriptor.file:
                            if m.output.self_import.get_package() != 'google.protobuf':
                                import_set.add(m.output.self_import.get_render_import())
                            else:
                                import_set.add(f'from google.protobuf import {m.output.self_import.get_class_name().lower()}_pb2')

        if import_set:
            return '\n'.join(sorted(import_set))
        return ''

    def get_template_context(self) -> Dict:
        d = super().get_template_context()
        d.update({
            'api_import': f'from {self.module.get_package()} import {self.module.get_module_name()[:-3]}api as api',
            'services': self.render_services(),
            'imports': self.render_imports(),
            'all_list': [f'{svc.service_descriptor.name}GrpcServicer' for svc in self.module.service_map.values()],
        })
        return d

    def render(self) -> str:
        return TemplateRenderer().render_file('grpc_receiver', **self.get_template_context())


class ServiceBuilder(base.AbstractServiceBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def render(self) -> str:
        __ = self.indent_str
        i = self.base_indent
        return (
            f'{i}class {self.service.service_descriptor.name}GrpcServicer(plasm.BaseGrpcServicer, pb2_grpc.{self.service.service_descriptor.name}Servicer):\n'
            f"{i}{__}def __init__(self, implementation: 'api.{self.service.service_descriptor.name}Interface'):\n"
            f'{i}{__}{__}super().__init__(implementation)\n'
            '\n'
            f'{i}{__}def add_to_server(self, server):\n'
            f'{i}{__}{__}pb2_grpc.add_{self.service.service_descriptor.name}Servicer_to_server(self, server)\n'
            '\n'
            f'{self.render_methods()}'
            '\n'
        )


class MethodBuilder(base.AbstractMethodBuilder):
    def render(self) -> str:
        __ = self.indent_str
        i = self.base_indent

        # TODO(thordurm@ccpgames.com>): The Request and Response may not be from the same file as the service!
        input_param_name = 'request_iterator' if self.method.io_type.is_input_stream() else 'request'
        return (
            f'{i}def {self.method.method_descriptor.name}(self, {input_param_name}: {self.method.get_request_type_hint()}, context: \'ServicerContext\') -> {self.method.get_response_type_hint()}:\n'
            f"{i}{__}return self.{self._forward_to_x_y_impl()}({self.method.get_response_dc_class_type()}, self.impl.{self.method.under_name}, {input_param_name}, context)\n"
        )

    def _forward_to_x_y_impl(self):
        return '_'.join([
            '_forward_to',
            'stream' if self.method.io_type.is_input_stream() else 'unary',
            'stream' if self.method.io_type.is_output_stream() else 'unary',
            'impl',
        ])



__all__ = [
    'ProtoType',
    'PyType',
    'PY_TYPE_MAP',
]
from google.protobuf.json_format import descriptor
from typing import *
import enum


class ProtoType(enum.IntEnum):
    UNKNOWN = -42
    DOUBLE = descriptor.FieldDescriptor.TYPE_DOUBLE  # 1
    FLOAT = descriptor.FieldDescriptor.TYPE_FLOAT  # 2
    INT64 = descriptor.FieldDescriptor.TYPE_INT64  # 3
    UINT64 = descriptor.FieldDescriptor.TYPE_UINT64  # 4
    INT32 = descriptor.FieldDescriptor.TYPE_INT32  # 5
    FIXED64 = descriptor.FieldDescriptor.TYPE_FIXED64  # 6
    FIXED32 = descriptor.FieldDescriptor.TYPE_FIXED32  # 7
    BOOL = descriptor.FieldDescriptor.TYPE_BOOL  # 8
    STRING = descriptor.FieldDescriptor.TYPE_STRING  # 9
    BYTES = descriptor.FieldDescriptor.TYPE_BYTES  # 12
    UINT32 = descriptor.FieldDescriptor.TYPE_UINT32  # 13
    SFIXED32 = descriptor.FieldDescriptor.TYPE_SFIXED32  # 15
    SFIXED64 = descriptor.FieldDescriptor.TYPE_SFIXED64  # 16
    SINT32 = descriptor.FieldDescriptor.TYPE_SINT32  # 17
    SINT64 = descriptor.FieldDescriptor.TYPE_SINT64  # 18

    GROUP = descriptor.FieldDescriptor.TYPE_GROUP  # 10
    MESSAGE = descriptor.FieldDescriptor.TYPE_MESSAGE  # 11
    ENUM = descriptor.FieldDescriptor.TYPE_ENUM  # 14

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.name}({self.value})'


class PyType:
    __slots__ = ('proto_type', 'py_name', 'default_value')

    def __init__(self, proto_type: Union[ProtoType, int], py_name: str, default_value: str):
        if isinstance(proto_type, int):
            proto_type = ProtoType(proto_type)
        self.proto_type = proto_type
        self.py_name = py_name
        self.default_value = default_value

    def __str__(self) -> str:
        return self.py_name

    def __repr__(self) -> str:
        if self.proto_type == ProtoType.UNKNOWN:
            return '?'
        return f'{self.proto_type!s}->{self.py_name}'

    # For backwards compatibility
    @property
    def number(self) -> int:
        return self.proto_type.value

    # For backwards compatibility
    @property
    def name(self) -> str:
        return self.py_name


PY_TYPE_MAP = {
    ProtoType.DOUBLE: PyType(ProtoType.DOUBLE, 'float', '0.0'),
    ProtoType.FLOAT: PyType(ProtoType.FLOAT, 'float', '0.0'),
    ProtoType.INT64: PyType(ProtoType.INT64, 'int', '0'),
    ProtoType.UINT64: PyType(ProtoType.UINT64, 'int', '0'),
    ProtoType.INT32: PyType(ProtoType.INT32, 'int', '0'),
    ProtoType.FIXED64: PyType(ProtoType.FIXED64, 'int', '0'),
    ProtoType.FIXED32: PyType(ProtoType.FIXED32, 'int', '0'),
    ProtoType.BOOL: PyType(ProtoType.BOOL, 'bool', 'False'),
    ProtoType.STRING: PyType(ProtoType.STRING, 'str', "''"),
    ProtoType.BYTES: PyType(ProtoType.BYTES, 'bytes', "b''"),
    ProtoType.UINT32: PyType(ProtoType.UINT32, 'int', '0'),
    ProtoType.SFIXED32: PyType(ProtoType.SFIXED32, 'int', '0'),
    ProtoType.SFIXED64: PyType(ProtoType.SFIXED64, 'int', '0'),
    ProtoType.SINT32: PyType(ProtoType.SINT32, 'int', '0'),
    ProtoType.SINT64: PyType(ProtoType.SINT64, 'int', '0'),

    # PyType.GROUP: PyType(PyType.GROUP, 'Group'),  # ?!?
    ProtoType.MESSAGE: PyType(ProtoType.MESSAGE, 'Message', 'None'),
    ProtoType.ENUM: PyType(ProtoType.ENUM, 'Enum', '0'),

    ProtoType.UNKNOWN: PyType(ProtoType.UNKNOWN, '?', 'None'),
}



from __future__ import annotations
from . import base
from ccptools.tpu import strimp
import typing

import logging
log = logging.getLogger(__name__)

_CACHE = {}


def get_builder_module(implementation_name: str):
    if implementation_name not in _CACHE:
        _CACHE[implementation_name] = strimp.get_module(f'{__package__}.{implementation_name}', logger=log, reraise=True)
    return _CACHE[implementation_name]


def get_module_builder(implementation_name: str) -> typing.Type[base.AbstractModuleBuilder]:
    return getattr(get_builder_module(implementation_name), 'ModuleBuilder')


def get_service_builder(implementation_name: str) -> typing.Type[base.AbstractServiceBuilder]:
    return getattr(get_builder_module(implementation_name), 'ServiceBuilder')


def get_method_builder(implementation_name: str) -> typing.Type[base.AbstractMethodBuilder]:
    return getattr(get_builder_module(implementation_name), 'MethodBuilder')



import unittest

from neobuilder.neobuilder import NeoBuilder
import os
import shutil

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)
import time

HERE = os.path.dirname(__file__)

PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')
EXPECTED_ROOT = os.path.join(HERE, 'res', 'expected')

EXPECTED_NUMBER_OF_FILES_CHECKED = 38

from neobuilder import __version__ as neobuilder_version
from protoplasm import __version__ as protoplasm_version


class NeobuilderTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

    def test_files_were_built(self):
        def assert_exists(file_to_check):
            self.assertTrue(os.path.exists(os.path.join(BUILD_ROOT, file_to_check)), msg=f'{file_to_check} not found')

        for (dirpath, dirnames, filenames) in os.walk(EXPECTED_ROOT):
            for f in filenames:
                assert_exists(os.path.join(dirpath, f))

    def test_code_in_files(self):
        file_count = 0
        for (dirpath, dirnames, filenames) in os.walk(EXPECTED_ROOT):
            for f in filenames:
                if not (f.endswith('_pb2.py') or f.endswith('_pb2_grpc.py')):  # We don't test the pb2 files!
                    rel_f = os.path.relpath(os.path.join(dirpath, f), EXPECTED_ROOT)  # noqa
                    file_count += 1

                    file_to_check = rel_f.replace('\\', '/')
                    lines_to_check = []
                    lines_skipped = 0
                    with open(os.path.join(BUILD_ROOT, file_to_check), 'r') as fin:  # noqa
                        for line in fin:
                            if not line.strip().startswith('# '):
                                stripped_line = line.rstrip()
                                if stripped_line:
                                    lines_to_check.append(stripped_line)
                            else:
                                lines_skipped += 1

                    lines_should_be = []
                    with open(os.path.join(EXPECTED_ROOT, file_to_check), 'r') as fin:  # noqa
                        for line in fin:
                            if not line.strip().startswith('# '):
                                if file_to_check == 'sandbox/__init__.py':
                                    if line.startswith("__protoplasm_version__ = '"):
                                        line = f"__protoplasm_version__ = '{protoplasm_version}'"
                                    if line.startswith("__neobuilder_version__ = '"):
                                        line = f"__neobuilder_version__ = '{neobuilder_version}'"
                                stripped_line = line.rstrip()
                                if stripped_line:
                                    lines_should_be.append(stripped_line)

                    line_count_should_be = len(lines_should_be)
                    line_count_is = len(lines_to_check)
                    with self.subTest(line_count_should_be=line_count_should_be, line_count_is=line_count_is, file_to_check=file_to_check):
                        self.assertEqual(line_count_should_be, line_count_is,
                                         msg=f'Line number mismatch in {file_to_check}')

                    with self.subTest(join_comparison_of_file=file_to_check):
                        self.assertEqual('\n'.join(lines_should_be),
                                         '\n'.join(lines_to_check), msg=f'Content difference in {file_to_check}')

                    if line_count_is == line_count_should_be:
                        for i in range(len(lines_should_be)):
                            with self.subTest(line_no=i + lines_skipped + 1, file_to_check=file_to_check):
                                self.assertEqual(lines_should_be[i], lines_to_check[i],
                                                 msg=f'Line no {i + lines_skipped + 1} in {file_to_check} does not match!')

                    else:
                        lines_should_be_set = set(lines_should_be)
                        lines_are_set = set(lines_to_check)
                        missing_lines = lines_should_be_set - lines_are_set
                        extra_lines = lines_are_set - lines_should_be_set
                        log.warning(f'There are {len(missing_lines)} missing lines in {file_to_check}:')
                        for l in missing_lines:
                            log.warning(f' - "{l}"')
                        log.warning(f'There are {len(extra_lines)} extra lines in {file_to_check}:')
                        for l in extra_lines:
                            log.warning(f' + "{l}"')

        self.assertEqual(EXPECTED_NUMBER_OF_FILES_CHECKED, file_count, msg='Did not check the expected number of files!')



__all__ = [
    'MethodDescriptorWrapper',
]

import dataclasses

from neobuilder.structs import *
from google.protobuf.json_format import descriptor
from protoplasm.casting import castutils


@dataclasses.dataclass
class MethodDescriptorWrapper:
    """This is a wrapper around Google's protobuf MethodDescriptor.

    Its point is twofold.

    1. To mask complex lookups and evaluations behind a simpler facade to use
       in the rest of the code.
    2. To abstract away the exact structure of Google's descriptor to guard
       against future code changes, especially in internal parts that aren't
       part of the official API.
    """
    real_descriptor: descriptor.MethodDescriptor = dataclasses.field(repr=False)

    name: str = dataclasses.field(default='', init=False)
    under_name: str = dataclasses.field(default='', init=False)

    client_streaming: bool = False
    server_streaming: bool = False

    input_type: Optional[descriptor.Descriptor] = dataclasses.field(default=None, init=False, repr=False)
    output_type: Optional[descriptor.Descriptor] = dataclasses.field(default=None, init=False, repr=False)

    io_type: MethodIoType = dataclasses.field(default=MethodIoType._UNKNOWN, init=False)

    service_descriptor: descriptor.ServiceDescriptor = dataclasses.field(default=None, init=False, repr=False)
    file_descriptor: descriptor.FileDescriptor = dataclasses.field(default=None, init=False, repr=False)

    def __post_init__(self):
        self.name = self.real_descriptor.name
        self.under_name = castutils.humps_to_under(self.name)
        self.input_type = self.real_descriptor.input_type
        self.output_type = self.real_descriptor.output_type
        self.client_streaming = self.real_descriptor.client_streaming
        self.server_streaming = self.real_descriptor.server_streaming
        if self.client_streaming:
            self.io_type |= MethodIoType._INPUT_STREAM
        else:
            self.io_type |= MethodIoType._INPUT_UNARY
        if self.server_streaming:
            self.io_type |= MethodIoType._OUTPUT_STREAM
        else:
            self.io_type |= MethodIoType._OUTPUT_UNARY

        self.service_descriptor = self.real_descriptor.containing_service
        self.file_descriptor = self.service_descriptor.file



__all__ = [
    'FieldKind',
    'FieldDescriptorWrapper',
]
import dataclasses
import enum
from typing import *
import keyword
from google.protobuf.json_format import descriptor
from google.protobuf import json_format
from neobuilder.structs.pytype import *


class FieldKind(enum.IntFlag):
    UNKNOWN = 0x0000
    DATA_SCALAR = 0x0001
    DATA_MESSAGE = 0x0002
    DATA_ENUM = 0x0004

    KIND_SINGLE = 0x0010
    KIND_LIST = 0x0020
    KIND_MAP = 0x0040

    SPECIAL_TIMESTAMP = 0x0100
    SPECIAL_BYTES = 0x0200
    SPECIAL_LONG = 0x0400
    SPECIAL_DURATION = 0x0800
    SPECIAL_ANY = 0x1000
    SPECIAL_EMPTY = 0x2000

    SPECIAL_KEYWORD = 0x10000000

    # scalar
    FIELD_SIMPLE = DATA_SCALAR | KIND_SINGLE
    # Message
    FIELD_MESSAGE = DATA_MESSAGE | KIND_SINGLE
    # Enum
    FIELD_ENUM = DATA_ENUM | KIND_SINGLE

    # repeated scalar
    FIELD_SIMPLE_LIST = DATA_SCALAR | KIND_LIST
    # repeated Message
    FIELD_MESSAGE_LIST = DATA_MESSAGE | KIND_LIST
    # repeated Enum
    FIELD_ENUM_LIST = DATA_ENUM | KIND_LIST

    # map<scalar, scalar>
    FIELD_SIMPLE_MAP = DATA_SCALAR | KIND_MAP
    # map<scalar, Message>
    FIELD_MESSAGE_MAP = DATA_MESSAGE | KIND_MAP
    # map<scalar, Enum>
    FIELD_ENUM_MAP = DATA_ENUM | KIND_MAP

    def __repr__(self) -> str:
        return self.name


@dataclasses.dataclass
class FieldDescriptorWrapper:
    """This is a wrapper around Google's protobuf FieldDescriptor.

    Its point is twofold.

    1. To mask complex lookups and evaluations behind a simpler facade to use
       in the rest of the code.
    2. To abstract away the exact structure of Google's descriptor to guard
       against future code changes, especially in internal parts that aren't
       part of the official API.
    """
    real_descriptor: descriptor.FieldDescriptor = dataclasses.field(repr=False)

    name: str = dataclasses.field(default='', init=False)

    kind: FieldKind = dataclasses.field(default=FieldKind.UNKNOWN, init=False)

    options: Optional[Any] = dataclasses.field(default=None, init=False, repr=False)

    key_field: Optional[descriptor.FieldDescriptor] = dataclasses.field(default=None, init=False, repr=False)
    value_field: Optional[descriptor.FieldDescriptor] = dataclasses.field(default=None, init=False, repr=False)
    value_msg: Optional[descriptor.Descriptor] = dataclasses.field(default=None, init=False, repr=False)

    py_type: PyType = dataclasses.field(default=PY_TYPE_MAP[ProtoType.UNKNOWN], init=False)
    key_py_type: PyType = dataclasses.field(default=PY_TYPE_MAP[ProtoType.UNKNOWN], init=False)

    is_proto_repeated: bool = dataclasses.field(default=False, init=False)

    def __post_init__(self):
        self.name = self.real_descriptor.name
        self.is_proto_repeated = self.real_descriptor.label == descriptor.FieldDescriptor.LABEL_REPEATED
        self._check_field_kind()

    def _load_options(self):
        self.options = None
        if self.real_descriptor.type == ProtoType.MESSAGE:
            if self.real_descriptor.message_type.has_options:
                self.options = self.real_descriptor.message_type.GetOptions()

    def _check_field_kind(self):
        self._load_options()

        if self.options and self.options.map_entry:  # Message that is a map...
            self.key_field = self.real_descriptor.message_type.fields_by_name['key']
            self.key_py_type = PY_TYPE_MAP[self.key_field.type]
            self.value_field = self.real_descriptor.message_type.fields_by_name['value']
            self.kind |= FieldKind.KIND_MAP
        else:
            self.value_field = self.real_descriptor
            if self.is_proto_repeated:
                self.kind |= FieldKind.KIND_LIST
            else:
                self.kind |= FieldKind.KIND_SINGLE

        if self.value_field.type == ProtoType.MESSAGE:  # Message
            self.value_msg = self.value_field.message_type

            if self.value_msg.full_name == 'google.protobuf.Timestamp':
                self.kind |= FieldKind.DATA_SCALAR
                self.kind |= FieldKind.SPECIAL_TIMESTAMP
            elif self.value_msg.full_name == 'google.protobuf.Duration':
                self.kind |= FieldKind.DATA_SCALAR
                self.kind |= FieldKind.SPECIAL_DURATION
            elif self.value_msg.full_name == 'google.protobuf.Any':
                self.kind |= FieldKind.DATA_SCALAR
                self.kind |= FieldKind.SPECIAL_ANY
            elif self.value_msg.full_name == 'google.protobuf.Empty':
                self.kind |= FieldKind.SPECIAL_EMPTY
            else:
                self.kind |= FieldKind.DATA_MESSAGE
        elif self.value_field.type == ProtoType.ENUM:  # Enum
            self.kind |= FieldKind.DATA_ENUM

        else:  # Scalar
            self.kind |= FieldKind.DATA_SCALAR

            if self.value_field.type == ProtoType.BYTES:
                self.kind |= FieldKind.SPECIAL_BYTES
            else:
                if self.value_field.cpp_type in json_format._INT64_TYPES:
                    self.kind |= FieldKind.SPECIAL_LONG

        self.py_type = PY_TYPE_MAP[self.value_field.type]

        if keyword.iskeyword(self.name):
            self.kind |= FieldKind.SPECIAL_KEYWORD

    @property
    def camelcase_name(self) -> str:
        return self.real_descriptor.camelcase_name

    @property
    def json_name(self) -> str:
        return self.real_descriptor.json_name

    @property
    def default_value(self) -> Any:
        return self.real_descriptor.default_value

    @property
    def full_name(self) -> Any:
        return self.real_descriptor.full_name


# def main():
#     fields = []
#     from sandbox.test import RainbowMessage
#     for f in RainbowMessage.DESCRIPTOR.fields:
#         ff = FieldDescriptorWrapper(f)
#         print(ff)
#         fields.append(ff)
#
#     # print(fields)
#
# if __name__ == '__main__':
#     main()



from . import base
import datetime


class ModuleBuilder(base.AbstractModuleBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_render_filename(self) -> str:
        return f'{self.module.get_file_full_name()[:-7]}_grpc_server.py'

    def render_top(self):
        return ('# Auto-Generated example file - Fill this out at will! :D\n'
                f'# Source module: {self.module.get_module_full_name()}\n'
                f'# Generated at: {datetime.datetime.now().isoformat()}\n')

    def render_services(self) -> str:
        return '\n'.join(self.render_service_list())

    def render(self):
        __ = self.indent_str
        return (
            f'{self.render_top()}'
            'from protoplasm import grpcserver\n'
            f'from {self.module.get_package()} import {self.module.get_module_name()[:-3]}grpc_receiver as grpc_receiver\n'
            f'from {self.module.get_package()} import {self.module.get_module_name()[:-3]}impl as impl\n'
            '\n'
            'import logging\n'
            'log = logging.getLogger(__name__)\n'
            '\n'
            '\n'
            f'def get_grpc_server() -> grpcserver.GrpcServer:\n'
            f'{__}gs = grpcserver.GrpcServer()\n'
            f'{self.render_services()}\n'
            f'{__}return gs\n'
            '\n'
            '\n'
            f"if __name__ == '__main__':\n"
            f'{__}logging.basicConfig(level=logging.DEBUG)\n'
            f'{__}gs = get_grpc_server()\n'
            f'{__}gs.serve()\n'
        )


class ServiceBuilder(base.AbstractServiceBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def render(self) -> str:
        __ = self.indent_str
        i = self.base_indent
        return f'{i}{__}gs.add_servicer(grpc_receiver.{self.service.service_descriptor.name}GrpcServicer(impl.{self.service.service_descriptor.name}()))'


class MethodBuilder(base.AbstractMethodBuilder):
    pass  # Not used!



from __future__ import annotations
from neobuilder.generators.symbols import service as symbols_service
from neobuilder.generators.symbols import modules
import datetime
from typing import *


class AbstractModuleBuilder(object):
    builder_name = ''

    def __init__(self, module: modules.ProtoModule):
        self.module = module
        self.indent_str = self.module.indent
        self.service_joiner = '\n\n'

    def get_template_context(self) -> Dict:
        return {
            'module_full_name': self.module.get_module_full_name(),
            'iso_timestamp': datetime.datetime.now().isoformat(),
        }

    def render_top(self) -> str:
        # TODO(thordurm@ccpgames.com) 2022-06-24: Remove... why?
        return ('# Auto-Generated file - DO NOT EDIT!\n'
                f'# Source module: {self.module.get_module_full_name()}\n'
                f'# Generated at: {datetime.datetime.now().isoformat()}\n')

    def render_service_list(self) -> List[str]:
        return [svc.render(self.builder_name) for svc in self.module.service_map.values()]

    def render_services(self) -> str:
        return '\n'.join(self.render_service_list())

    def get_render_filename(self) -> str:
        raise NotImplementedError('AbstractModuleBuilder not implemented')

    def render(self) -> str:
        raise NotImplementedError('AbstractModuleBuilder not implemented')


class AbstractServiceBuilder(object):
    builder_name = ''

    def __init__(self, service: symbols_service.ProtoService, indent_level: int = 0):
        self.service = service
        self.indent_level = indent_level
        self.indent_str = self.service.indent
        self.base_indent = self.indent_str * indent_level

    def get_template_context(self) -> Dict:
        return {
            'is_grpc': self.service.module.is_grpc_file(),
            'service': self.service,
            'service_name': self.service.service_descriptor.name,
            '_indent_level': self.indent_level,
            '_indent_str': self.indent_str,
            '_base_indent': self.base_indent,
            'method_list': [self.get_method_context(m) for m in self.service.method_map.values()],
            'indent_spaces': self.indent_level * 4,
        }

    def get_method_context(self, method: symbols_service.ProtoMethod) -> Dict[str, Any]:
        return {
            'method': method,
        }

    def render_methods(self) -> str:
        ____ = self.indent_str
        i = self.base_indent
        buf = []
        if self.service.method_map:
            buf.append(
                '\n'.join([m.render(self.builder_name, self.indent_level + 1) for m in self.service.method_map.values()]))
        else:
            buf.append(f'{i}pass\n')
        return ''.join(buf)

    def render(self) -> str:
        raise NotImplementedError('AbstractServiceBuilder not implemented')


class AbstractMethodBuilder(object):
    def __init__(self, method: symbols_service.ProtoMethod, indent_level: int = 0):
        self.method = method
        self.indent_level = indent_level
        self.indent_str = self.method.indent
        self.base_indent = self.indent_str * indent_level

    def get_template_context(self) -> Dict:
        return {
            'method': self.method,
            '_indent_level': self.indent_level,
            '_indent_str': self.indent_str,
            '_base_indent': self.base_indent,
            'indent_spaces': self.indent_level * 4,
        }

    def render(self) -> str:
        raise NotImplementedError('AbstractServiceMethod not implemented')



from __future__ import annotations
__all__ = [
    'ProtoService',
    'ProtoMethod',
]
from google.protobuf import json_format

from neobuilder.descwrap import *
from neobuilder.structs import *

from neobuilder.generators import servicebuilders
from neobuilder.generators.symbols import dataclass
from neobuilder.generators.symbols import modules


class ProtoService(object):
    def __init__(self, service_descriptor: json_format.descriptor.ServiceDescriptor, module: modules.ProtoModule):
        self.service_descriptor = service_descriptor
        # self.service_descriptor.name = Math
        # self.service_descriptor.file.name = sandbox/test/service.proto
        # self.service_descriptor.file.package = sandbox.test
        # self.service_descriptor.full_name = sandbox.test.Math
        self.module = module
        self.method_map: Mapping[str, ProtoMethod] = {}
        self.indent = module.indent
        self._load_service()

    def _load_service(self):
        self.method_map = {m.name: ProtoMethod(m, self) for m in self.service_descriptor.methods}

    # ---

    def render_grpc_server(self, indent_level=0):  # GRPC Server
        return self.render('grpc_server', indent_level)
        # return f'{self.indent*indent_level}gs.add_servicer(dc_grpc.{self.service_descriptor.name}GrpcServicer(impl.{self.service_descriptor.name}()))'

    def render_api_service(self, indent_level=0):  # API
        return self.render('interface', indent_level)

    def render_grpc_servicer(self, indent_level=0):  # GRPC DC
        return self.render('grpc_receiver', indent_level)

    def render_grpc_impl(self, indent_level=0):  # GRPC Implementation
        return self.render('grpc_sender', indent_level)

    def render_impl_service(self, indent_level=0):  # Implementation example
        return self.render('implementation', indent_level)

    #
    # Generic renderer
    #

    def render(self, builder_name: str, indent_level: int = 0) -> str:
        return servicebuilders.get_service_builder(builder_name)(self, indent_level).render()


class ProtoMethod(object):
    def __init__(self,
                 method_descriptor: json_format.descriptor.MethodDescriptor,
                 service: ProtoService):
        self.method_descriptor: MethodDescriptorWrapper = MethodDescriptorWrapper(method_descriptor)
        # self.method_descriptor.name = Add
        # self.method_descriptor.full_name = sandbox.test.Math.Add
        self.service = service
        self.indent = service.indent

        self.input = dataclass.ProtoClass(self.method_descriptor.input_type, self.service.module, force_import=True)
        self.output = dataclass.ProtoClass(self.method_descriptor.output_type, self.service.module, force_import=True)

    @property
    def under_name(self) -> str:
        return self.method_descriptor.under_name

    def get_return_type_hint(self, relative_to_self: bool = False):
        if self.io_type.is_output_stream():
            if relative_to_self:
                return f'plasm.ResponseIterator[{self.output.render_api_return_dataclass(self.service.service_descriptor.file)}]'
            return f'plasm.ResponseIterator[{self.output.render_api_return_dataclass()}]'
        if relative_to_self:
            return self.output.render_api_return_args(self.service.service_descriptor.file)
        return self.output.render_api_return_args()

    def get_input_params(self, relative_to_self: bool = False):
        # TODO(thordurm@ccpgames.com) 2022-06-18: Check for input stream...
        if self.io_type.is_input_stream():
            return self.input.render_api_request_arg(self.method_descriptor)

        else:
            if relative_to_self:
                args = self.input.render_api_args(self.service.service_descriptor.file)
            else:
                args = self.input.render_api_args()
            if args:
                return f', {args}'
            else:
                return ''

    def get_fwd_params(self):
        if self.io_type.is_input_stream():
            return ', request_iterator'

        args = self.input.render_forwarded_args()
        if args:
            return f', {args}'
        else:
            return ''

    def get_input_name(self):
        return f'{self.input.get_class_name()}'

    def get_output_name(self):
        return f'{self.output.get_class_name()}'

    def get_request_type_hint(self) -> str:
        unary = f'{self.input.render_pb2_type_hint(self.service.module)}'
        if self.io_type.is_input_stream():
            return f'Iterable[{unary}]'
        return unary

    def get_request_dc_type_hint(self) -> str:
        return f'{self.input.render_dc_type_hint(self.service.module)}'

    def get_response_type_hint(self) -> str:
        return f'{self.output.render_pb2_type_hint(self.service.module)}'

    def get_response_dc_type(self) -> str:
        if self.io_type.is_output_stream():
            return f'Iterable[{self.get_response_dc_class_type()}]'
        return self.get_response_dc_class_type()

    def get_response_dc_class_type(self) -> str:
        return f'{self.output.render_dc_type_hint(self.service.module)}'

    def render(self, generator_name: str, indent_level: int = 0) -> str:
        return servicebuilders.get_method_builder(generator_name)(self, indent_level).render()

    @property
    def io_type(self) -> MethodIoType:
        return self.method_descriptor.io_type



from __future__ import annotations

__all__ = [
    'PyType',
    'PY_TYPE_MAP',
    'ProtoImport',
]

import os
import typing

from google.protobuf import json_format


from neobuilder.structs.pytype import *


class ProtoImport(object):
    def __init__(self, descriptor: typing.Union[json_format.descriptor.Descriptor, json_format.descriptor.EnumDescriptor], is_enum=False):
        self.descriptor = descriptor
        # noinspection PyUnresolvedReferences
        self.file_descriptor: json_format.descriptor.FileDescriptor = descriptor.file
        self.is_enum = is_enum

        self.name = self.descriptor.name

        self.module = f'{self.file_descriptor.name[:-6].replace("/", ".")}_dc'

    def get_full_class_name(self):
        return f'{self.get_module_full_name()}.{self.get_class_name()}'

    def get_class_name(self):
        return self.name

    def get_module_full_name(self):
        return self.module

    def get_module_name(self):
        return self.module.split('.')[-1]

    def get_module_name_parts(self):
        return self.module.split('.')

    def get_package(self):
        return self.module.rsplit('.', 1)[0]

    def get_file_full_name(self):
        return f"{self.module.replace('.', '/')}.py"

    def get_file_name(self):
        return os.path.basename(self.get_file_full_name())

    def get_file_path(self):
        return os.path.dirname(self.get_file_full_name())

    def get_name(self):
        return os.path.splitext(self.get_file_name())[0]

    def get_render_file_name(self):
        return f'{self.get_file_full_name()[:-7]}_dc.py'

    def get_import_name(self, used_in_file: json_format.descriptor.FileDescriptor = None):
        if used_in_file and used_in_file == self.descriptor.file:
            return 'dc'
        return self.get_module_full_name().replace('.', '__')

    def get_render_import(self, used_in_file: json_format.descriptor.FileDescriptor = None):
        return f'from {self.get_package()} import {self.get_module_name()} as {self.get_import_name(used_in_file)}'



import argparse

from neobuilder.neobuilder import NeoBuilder


def main():
    parser = argparse.ArgumentParser(description='Builds neobuf packages with protoplasm.',
                                     epilog=(f'Neobuilder v{NeoBuilder.neobuilder_version()} - '
                                             f'Protoplasm v{NeoBuilder.protoplasm_version()}'))
    version_group = parser.add_mutually_exclusive_group()
    version_group.add_argument('-m', '--major', action="store_true",
                               help='Bump the major version number instead of the minor')
    version_group.add_argument('-p', '--patch', action="store_true",
                               help='Bump the patch version number instead of the minor')
    parser.add_argument('package', help='Package name')
    parser.add_argument('protopath', help='Path to the root of the protobuf files (default="./proto")',
                        default='./proto', nargs='?')
    parser.add_argument('-b', '--buildroot', help='Path to the root of the output build files (default="./build")',
                        default='./build', nargs='?')
    parser.add_argument('-v', '--verbose', action="store_true",
                        help='Spits out DEBUG level logs')

    args = parser.parse_args()

    n = NeoBuilder(
        package=args.package,
        protopath=args.protopath,
        build_root=args.buildroot,
        major=args.major,
        patch=args.patch,
        verbose=args.verbose,
    )
    n.build()


if __name__ == '__main__':
    main()



from . import base
import datetime


class ModuleBuilder(base.AbstractModuleBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_render_filename(self) -> str:
        return f'{self.module.get_file_full_name()[:-7]}_impl.py'

    def render_top(self):
        return ('# Auto-Generated example file - Fill this out at will! :D\n'
                f'# Source module: {self.module.get_module_full_name()}\n'
                f'# Generated at: {datetime.datetime.now().isoformat()}\n')

    def render(self) -> str:
        return (
            f'{self.render_top()}'
            f'from {self.module.get_package()}.{self.module.get_module_name()[:-3]}api import *\n'
            'from protoplasm import decorators\n'
            '\n'
            'import logging\n'
            'log = logging.getLogger(__name__)\n'
            '\n'
            '\n'
            f'{self.render_services()}'
        )


class ServiceBuilder(base.AbstractServiceBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def render(self) -> str:
        ____ = self.indent_str
        i = self.base_indent
        return (
            f'{i}class {self.service.service_descriptor.name}({self.service.service_descriptor.name}Interface):\n'
            f'{self.render_methods()}'
        )


class MethodBuilder(base.AbstractMethodBuilder):
    def render(self) -> str:
        ____ = self.indent_str
        i = self.base_indent
        return (
            f'{i}def {self.method.under_name}(self{self.method.get_input_params()}) -> {self.method.get_return_type_hint()}:\n'
            f"{i}{____}raise errors.api.Unimplemented()\n"
        )



__all__ = [
    'NeoBuilder',
]


import sys
import os
import shutil
import time
from importlib import resources
import semver

from grpc_tools import protoc
from . import render
import pathlib

from ccptools.tpu import strimp

from neobuilder.generators.symbols import modules
from neobuilder.structs import *
from neobuilder import __version__ as neobuilder_version

import logging
log = logging.getLogger(__name__)

_PY_3_9_PLUS = sys.version_info >= (3, 9)


class NeoBuilder(object):
    VERSION = neobuilder_version

    def __init__(self,
                 package: str,
                 protopath: str = './proto',
                 build_root: str = './build',
                 major: bool = False,
                 patch: bool = False,
                 verbose: bool = False):

        self.package = package

        self.protopath = protopath.replace('\\', '/')
        self.build_root = build_root.replace('\\', '/')

        self.major = major
        self.patch = patch

        self.proto_files: List[ProtoFile] = []

        # Proto Build Arguments and stuff...
        self.proto_include = self._get_basic_proto_path()
        self.proto_build_args = ['protoc', f'-I{self.protopath}/']
        if self.proto_include:
            self.proto_build_args.append('-I{}'.format(self.proto_include))
        self.proto_build_args.append(f'--python_out={self.build_root}')

        self.last_version = None
        self._next_version = None
        self.commit_message = 'No commit message'
        self.verbose = verbose
        logging.basicConfig(level=(logging.DEBUG if self.verbose else logging.INFO),
                            format='%(levelname)8s - %(message)s')
        log.info(f'Initializing Neobuilder {self.neobuilder_version()}')

        log.debug(f'{self.package=}')
        log.debug(f'{self.protopath=}')
        log.debug(f'{self.build_root=}')
        log.debug(f'{self.proto_include=}')

    @staticmethod
    def _get_basic_proto_path() -> str:
        if _PY_3_9_PLUS:
            from importlib import resources
            return str(resources.files('grpc_tools').joinpath('_proto'))
        else:
            import pkg_resources  # noqa
            return str(pkg_resources.resource_filename('grpc_tools', '_proto'))  # noqa

    @property
    def next_version(self):
        if self._next_version:
            return self._next_version
        return '0.1.0'

    @property
    def build_path(self) -> str:
        return os.path.join(self.build_root, self.package)

    def read_last_version(self):
        log.debug('Reading last version...')
        version_file = os.path.join(self.build_path, '__init__.py')
        if os.path.exists(version_file):
            try:
                m = strimp.get_module(self.package)
                version_var = getattr(m, '__version__', None)
                if not version_var:
                    log.warning('__version__ not found in package')

                else:
                    if isinstance(version_var, tuple):
                        if len(version_var) > 3:
                            ver_str = '.'.join(version_var[:3])
                        else:
                            ver_str = '.'.join(version_var)
                    else:
                        ver_str = version_var

                    self.last_version = ver_str
                    log.info(f'Last version: {self.last_version}')

            except Exception as ex:
                log.exception('package __init__ file import error: %r' % ex)

        else:
            log.warning('package __init__ file not found (is this a new/first build?)')

        if not self.last_version:
            log.warning('Assuming last version: 0.0.0.0')
            self.last_version = ''

    def read_next_version(self):
        log.debug('Reading next version...')
        version_file = os.path.join(self.protopath, self.package, 'VERSION')

        if os.path.exists(version_file):
            version_str = ''
            rest = []
            with open(version_file, 'r') as fin:
                first = True
                for line in fin:
                    if first:
                        version_str = line.strip()
                        first = False
                    else:
                        rest.append(line)

            if not version_str:
                log.warning('VERSION file empty!')

            else:
                try:
                    if '-' in version_str:
                        sv = semver.Version.parse(version_str)
                    else:
                        parts = version_str.split('.')
                        if len(parts) > 3:
                            sv = semver.Version.parse('.'.join(version_str[:3]))
                        else:
                            sv = semver.Version.parse(version_str)

                    self._next_version = str(sv)

                    if rest:
                        msg = '\n'.join(rest)
                        msg = msg.strip()
                        if msg:
                            self.commit_message = msg
                except Exception as ex:
                    log.exception('VERSION file parsing error: %r' % ex)

        else:
            log.warning('VERSION file not found!')

        if not self._next_version:
            log.info('Trying to calculate next version...')
            if not self.last_version:
                log.warning('No last version to calculate next version :(')

            self._next_version = self.calc_next_version()
        log.info(f'Next version: {self._next_version}')

    def calc_next_version(self):
        if self.last_version:
            this_ver = semver.Version.parse(self.last_version)
            if self.major:
                log.info('Bumping major version...')
                next_up = this_ver.bump_major()
            elif self.patch:
                log.info('Bumping patch version...')
                next_up = this_ver.bump_patch()
            else:
                log.info('Bumping minor version...')
                next_up = this_ver.bump_minor()
            return str(next_up)
        return '0.1.0'

    def collect_proto_files(self):
        log.debug('Collecting proto files...')

        def _grpc_check(filename):
            with open(filename, 'r') as fin:
                for line in fin:
                    if line.startswith('option py_generic_services = true'):
                        log.debug(f'Generic service... skip gRPC: {filename}')
                        return False
                    if line.startswith('service '):
                        log.debug(f'Service found: {filename}')
                        return True
            log.debug(f'Service not found: {filename}')
            return False

        self.proto_files = []
        walkroot = os.path.join(self.protopath, self.package)
        rel_root = os.path.join(self.protopath)
        for (dirpath, dirnames, filenames) in os.walk(walkroot):
            for f in filenames:
                if f.endswith('.proto'):
                    dirp = dirpath.replace('\\', '/')
                    rel_dir = dirp[len(rel_root):]
                    if rel_dir.startswith('/'):
                        rel_dir = rel_dir[1:]
                    if not rel_dir.startswith(self.package):
                        log.error(f'Error parsing relative path for {f}')
                    # log.debug(f'... {f=}')
                    # log.debug(f'... {rel_root=}')
                    # log.debug(f'... rel_root={rel_root!r}')
                    # log.debug(f'... {dirp=}')
                    # log.debug(f'... {rel_dir=}')
                    pf = ProtoFile(f, dirp, rel_dir)
                    pf.is_grpc_service = _grpc_check(pf.full_name())
                    self.proto_files.append(pf)
                    log.debug(f'Found proto file: {pf}')
        log.info(f'Found {len(self.proto_files)} proto files!')

    def delete_backup(self):
        backup_dir = f'{self.build_path}__backup'
        if os.path.exists(backup_dir):
            log.debug(f'Deleting backup dir: {backup_dir}')
            shutil.rmtree(backup_dir)
            time.sleep(0.1)

    def purge_old(self):
        backup_dir = f'{self.build_path}__backup'
        fail_dir = f'{self.build_path}__fail'
        if os.path.exists(fail_dir):
            log.debug(f'(purge old) Deleting old failed build dir: {fail_dir}')
            shutil.rmtree(fail_dir)
            time.sleep(0.1)

        if os.path.exists(self.build_path):
            if os.path.exists(backup_dir):
                log.debug(f'(purge old) Deleting backup dir: {fail_dir}')
                shutil.rmtree(backup_dir)
                time.sleep(0.1)

            log.debug(f'(purge old) Moving dir: {self.build_path} -> {backup_dir}')
            shutil.move(self.build_path, backup_dir)
            time.sleep(0.1)

    def revert_old(self):
        backup_dir = f'{self.build_path}__backup'
        fail_dir = f'{self.build_path}__fail'
        if os.path.exists(backup_dir) and os.path.exists(self.build_path):
            if os.path.exists(fail_dir):
                log.debug(f'(revert_old) Deleting old failed build: {fail_dir}')
                shutil.rmtree(fail_dir)
                time.sleep(0.1)
            log.debug(f'(revert_old) Moving dir: {self.build_path} -> {fail_dir}')
            shutil.move(self.build_path, fail_dir)
            time.sleep(0.1)
            log.debug(f'(revert_old) Moving dir: {backup_dir} -> {self.build_path}')
            shutil.move(backup_dir, self.build_path)
            time.sleep(0.1)

    def build(self):
        try:
            sys.path.extend([os.path.abspath(self.build_root)])
            log.debug(f'PATH={sys.path}')
            self.collect_proto_files()
            self.read_last_version()
            self.read_next_version()
            self.purge_old()
            log.info('Waiting 1 sec after IO operations before building proto files...')
            time.sleep(1)

            if not os.path.exists(self.build_path):
                log.debug(f'Generating build path: {self.build_path}')
                os.makedirs(self.build_path, exist_ok=True)

            try:
                log.info('Compiling protobuf files...')
                failed = []

                for pf in self.proto_files:
                    if not self.proto_build(pf):
                        failed.append(pf)

                if failed:
                    for f in failed:
                        log.error(f'Failed to compile proto file: {f}')
                    raise RuntimeError('failed to compile proto files')

            except Exception as ex:
                log.exception(ex)
                log.error('Reverting backup...')
                self.revert_old()
                self.fail_with_style(10, 'Failed to compile proto files with protoc')

            log.info('Waiting 1 sec after IO ops before building neobuf files...')
            time.sleep(1)
            log.info('Generating neobuf files...')
            try:
                for pf in self.proto_files:
                    self.plasm_build(pf)
            except Exception as ex:
                log.exception(ex)
                log.error('Reverting backup...')
                self.revert_old()
                self.fail_with_style(20, 'Failed to build neobuf files with protoplasm')

            log.info('Waiting 1 sec after IO ops before finalizing...')
            time.sleep(1)

            self.add_version()
            self.add_inits()
            self.add_everything()
            self.write_commit_file()
            self.delete_backup()
            log.info('Neobuilder Done!')
        except Exception as ex:
            log.exception(ex)
            self.fail_with_style(99, 'Got 99 problems and an unknown exception is one!')

    def proto_build(self, protofile: ProtoFile) -> bool:
        input_args = self.proto_build_args.copy()
        if protofile.is_grpc_service:
            input_args.append(f'--grpc_python_out={self.build_root}')
        input_args.append(protofile.full_name())
        log.debug(f'Executing: protoc {" ".join(input_args)}')

        ret_val = protoc.main(input_args)

        if ret_val != 0:
            return False
        return True

    def plasm_build(self, protofile: ProtoFile):
        log.info(f'Neobuffing {protofile}')
        m = strimp.get_module(protofile.pb2_module(), logger=log, reraise=True)
        if not m:
            log.error(f'protofile={protofile}')
            raise ImportError('ERROR! Module not found! %s' % protofile.pb2_module())
        else:
            try:
                p = modules.ProtoModule(m)
                log.debug(f' - Writing ProtoModule: {p.get_module_full_name()}')
                p.write_rendered_file()
            except Exception as ex:
                log.exception(f'ERROR! Bad stuff happened! %r' % ex)
                raise

    def add_inits(self):
        log.debug('Adding __init__ files to generated modules...')
        for dirpath, dirnames, filenames in os.walk(self.build_path):
            if not dirpath.endswith('__pycache__'):
                if '__init__.py' not in filenames:
                    render.to_file(f'{dirpath}/__init__.py', render.init())

    def add_version(self):
        log.debug('Adding version to generated package root...')
        package_init = os.path.join(self.build_path, '__init__.py')
        if not os.path.exists(package_init):
            render.to_file(package_init, render.root_init(
                package_name=self.package,
                version_str=self.next_version
            ))
        else:
            raise ValueError('Package __init__ file (%s) already exists? :(', package_init)

    def write_commit_file(self):
        log.debug('Writing commit file...')
        with open(os.path.join(self.build_path, 'VERSION'), 'w', encoding='utf-8') as fout:
            fout.write(f'Version {self.next_version}')
            if self.commit_message:
                fout.write(' - ')
                fout.write(self.commit_message)
            fout.write('\n')

    def add_everything(self):
        # TODO(thordurm@ccpgames.com) 2022-06-07: If we start using __all__ directives then this may well change...
        log.debug('Generating __everything__ file...')
        everything_file = os.path.join(self.build_path, '__everything__.py')
        counter = 0
        with open(everything_file, 'w', encoding='utf-8') as fout:
            fout.write('# Autogenerated file that imports every proto and protoplasm file.\n')
            fout.write('# This is to ensure that the proto symbol database gets populated with everything\n')
            fout.write('# which is something required for Any deserialization to work\n\n')

            line_buffer = []

            broot = pathlib.Path(self.build_root).absolute()

            for (dirpath, dirnames, filenames) in os.walk(self.build_path):
                for f in filenames:
                    if f.endswith('.py') and not f.endswith('__.py'):
                        fpath = pathlib.Path(dirpath) / f
                        rel_path = fpath.relative_to(broot)
                        package_name = str(rel_path)[:-3].replace('\\', '/').replace('/', '.')
                        line_buffer.append(f'import {package_name}')
                        counter += 1
            line_buffer.sort()
            fout.write('\n'.join(line_buffer))
            fout.write('\n')
        log.debug(f'Added {counter} imports to __everything__')

    @staticmethod
    def neobuilder_version() -> str:
        if isinstance(NeoBuilder.VERSION, tuple):
            return '.'.join([str(i) for i in NeoBuilder.VERSION])
        else:
            return NeoBuilder.VERSION

    @staticmethod
    def protoplasm_version() -> str:
        from protoplasm import __version__ as protoplasm_version
        if isinstance(protoplasm_version, tuple):
            return '.'.join([str(i) for i in protoplasm_version])
        else:
            return protoplasm_version

    @staticmethod
    def fail_with_style(code: int, message: str):
        log.error(message)
        sys.exit(code)



from . import base
import grpc
from typing import *

from ...neobuilder.render import TemplateRenderer


class ModuleBuilder(base.AbstractModuleBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def get_render_filename(self) -> str:
        return f'{self.module.get_file_full_name()[:-7]}_grpc_sender.py'

    def render_imports(self) -> str:
        import_set = {
            f'from {self.module.get_package()} import {self.module.get_module_name()}_grpc as pb2_grpc\n'
            f'from {self.module.get_package()}.{self.module.get_module_name()[:-3]}api import *\n'
        }

        for s in self.module.service_map.values():
            for m in s.method_map.values():
                if m.input:
                    if m.input.self_import and m.input.self_import.get_package() != 'google.protobuf':
                        import_set.add(m.input.self_import.get_render_import(self.module.descriptor))

        if import_set:
            return '\n'.join(sorted(import_set))
        return ''

    def get_template_context(self) -> Dict:
        d = super().get_template_context()
        d.update({
            'imports': self.render_imports(),
            'services': self.render_services(),
            'all_list': [f'{svc.service_descriptor.name}' for svc in self.module.service_map.values()],
        })
        return d

    def render(self) -> str:
        return TemplateRenderer().render_file('grpc_sender', **self.get_template_context())


class ServiceBuilder(base.AbstractServiceBuilder):
    builder_name = __name__.split('.')[-1]  # Grab module name

    def render(self) -> str:
        __ = self.indent_str
        i = self.base_indent
        return (
            f'{i}class {self.service.service_descriptor.name}(plasm.BaseGrpcClientImplementation, {self.service.service_descriptor.name}Interface):\n'
            f'{i}{__}def __init__(self, grpc_host: str = \'localhost:50051\', credentials: Optional[Union[bool, \'ChannelCredentials\']] = None, options: Optional[Dict] = None, *args, **kwargs):\n'
            f'{i}{__}{__}super().__init__(pb2_grpc.{self.service.service_descriptor.name}Stub, grpc_host, credentials, options, *args, **kwargs)\n'
            f'\n'
            f'{self.render_methods()}'
        )


class MethodBuilder(base.AbstractMethodBuilder):
    def render(self) -> str:
        __ = self.indent_str
        i = self.base_indent
        return (
            f'{i}def {self.method.under_name}(self{self.method.get_input_params(True)}) -> {self.method.get_return_type_hint(True)}:\n'
            f"{i}{__}return self._forward_to_grpc({self.method.get_request_dc_type_hint()}, self.stub.{self.method.method_descriptor.name}{self.method.get_fwd_params()})\n"
        )



__all__ = [
    'ProtoFile',
]

import dataclasses
import os


@dataclasses.dataclass
class ProtoFile:  # E.g. ./proto/ccpgears/enrichment/ebs/eve/character/character.proto
    name: str  # E.g. character.proto
    path: str  # E.g. ./proto/ccpgears/enrichment/ebs/eve/character/
    rel_path: str  # E.g. ccpgears/enrichment/ebs/eve/character/
    is_grpc_service: bool = dataclasses.field(default=False)

    def full_name(self) -> str:
        return os.path.join(self.path, self.name).replace('\\', '/')

    def pb2_module(self) -> str:
        return f'{os.path.join(self.rel_path, self.name)[:-6].replace("/", ".")}_pb2'.replace('\\', '.')



from .renderer import TemplateRenderer
from typing import *


def to_file(file: str, content: str):
    with open(file, 'w', encoding='utf-8') as fout:
        fout.write(content)
        fout.write('\n')


def init() -> str:
    return TemplateRenderer().render_file('init')


def root_init(package_name: str,
              version_str) -> str:
    from protoplasm import __version__ as ppv
    from neobuilder import __version__ as nbv
    if isinstance(ppv, tuple):
        if len(ppv) > 3:
            ppv = '.'.join(str(i) for i in ppv[:3])
        else:
            ppv = '.'.join(str(i) for i in ppv)
    if isinstance(nbv, tuple):
        if len(nbv) > 3:
            nbv = '.'.join(str(i) for i in nbv[:3])
        else:
            nbv = '.'.join(str(i) for i in nbv)

    return TemplateRenderer().render_file('root_init',
                                          version=version_str,
                                          protoplasm_version=ppv,
                                          neobuilder_version=nbv,
                                          package_name=package_name)





import unittest
import os
import sys
from neobuilder.neobuilder import NeoBuilder
import shutil
import time
import neobuilder.generators.symbols
import neobuilder.generators.symbols.dataclass
import neobuilder.generators.symbols.modules

HERE = os.path.dirname(__file__)
BUILD_ROOT = os.path.join(HERE, 'res', 'build')
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class ProtoFieldTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        sys.path.append(BUILD_ROOT)

    def test_rainbow_simple_field(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['simple_field'], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.field_descriptor.is_proto_repeated)

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('simple_field', proto_field.py_name())
        self.assertEqual('str', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator", proto_field.get_metadata())
        self.assertEqual("simple_field: str = dataclasses.field(default='', metadata={'dictator': dictators.BaseDictator})", proto_field.render_py())

    def test_rainbow_message_field(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['message_field'], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertTrue(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('message_field', proto_field.py_name())
        self.assertEqual('SubMessage', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator, 'is_obj': True", proto_field.get_metadata())
        self.assertEqual("message_field: SubMessage = dataclasses.field(default=None, metadata={'dictator': dictators.BaseDictator, 'is_obj': True})", proto_field.render_py())

    def test_rainbow_simple_list(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['simple_list'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('simple_list', proto_field.py_name())
        self.assertEqual('List[str]', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator, 'is_list': True", proto_field.get_metadata())
        self.assertEqual("simple_list: List[str] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})", proto_field.render_py())

    def test_rainbow_message_list(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['message_list'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertTrue(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('message_list', proto_field.py_name())
        self.assertEqual('List[SubMessage]', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator, 'is_list': True, 'is_obj': True", proto_field.get_metadata())
        self.assertEqual("message_list: List[SubMessage] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True, 'is_obj': True})", proto_field.render_py())

    def test_rainbow_simple_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['simple_map'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('simple_map', proto_field.py_name())
        self.assertEqual('Dict[str, str]', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator, 'is_map': True", proto_field.get_metadata())
        self.assertEqual("simple_map: Dict[str, str] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})", proto_field.render_py())

    def test_rainbow_message_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.RainbowMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['message_map'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertTrue(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('message_map', proto_field.py_name())
        self.assertEqual('Dict[str, SubMessage]', proto_field.get_type_hint())
        self.assertEqual('BaseDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True", proto_field.get_metadata())
        self.assertEqual("message_map: Dict[str, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})", proto_field.render_py())

    def test_timestamp(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.TimestampMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_timestamp'], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('my_timestamp', proto_field.py_name())
        self.assertEqual('datetime.datetime', proto_field.get_type_hint())
        self.assertEqual('TimestampDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.TimestampDictator", proto_field.get_metadata())
        self.assertEqual("my_timestamp: datetime.datetime = dataclasses.field(default=None, metadata={'dictator': dictators.TimestampDictator})", proto_field.render_py())

    def test_timestamp_list(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.TimestampMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_timestamp_list'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('my_timestamp_list', proto_field.py_name())
        self.assertEqual('List[datetime.datetime]', proto_field.get_type_hint())
        self.assertEqual('TimestampDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.TimestampDictator, 'is_list': True", proto_field.get_metadata())
        self.assertEqual("my_timestamp_list: List[datetime.datetime] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.TimestampDictator, 'is_list': True})", proto_field.render_py())

    def test_timestamp_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.TimestampMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_timestamp_map'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('my_timestamp_map', proto_field.py_name())
        self.assertEqual('Dict[str, datetime.datetime]', proto_field.get_type_hint())
        self.assertEqual('TimestampDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.TimestampDictator, 'is_map': True", proto_field.get_metadata())
        self.assertEqual("my_timestamp_map: Dict[str, datetime.datetime] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.TimestampDictator, 'is_map': True})", proto_field.render_py())

    def test_bytes(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.BytesMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_bytes'], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertTrue(proto_field.is_bytes())

        self.assertEqual('my_bytes', proto_field.py_name())
        self.assertEqual('bytes', proto_field.get_type_hint())
        self.assertEqual('ByteDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.ByteDictator", proto_field.get_metadata())
        self.assertEqual("my_bytes: bytes = dataclasses.field(default=b'', metadata={'dictator': dictators.ByteDictator})", proto_field.render_py())

    def test_bytes_list(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.BytesMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_bytes_list'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertTrue(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('my_bytes_list', proto_field.py_name())
        self.assertEqual('List[bytes]', proto_field.get_type_hint())
        self.assertEqual('ByteDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.ByteDictator, 'is_list': True", proto_field.get_metadata())
        self.assertEqual("my_bytes_list: List[bytes] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.ByteDictator, 'is_list': True})", proto_field.render_py())

    def test_bytes_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.BytesMessage.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name['my_bytes_map'], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertFalse(proto_field.is_timestamp())
        self.assertTrue(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual('my_bytes_map', proto_field.py_name())
        self.assertEqual('Dict[str, bytes]', proto_field.get_type_hint())
        self.assertEqual('ByteDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.ByteDictator, 'is_map': True", proto_field.get_metadata())
        self.assertEqual("my_bytes_map: Dict[str, bytes] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.ByteDictator, 'is_map': True})", proto_field.render_py())

    def test_all_types(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.AllTypes.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        class_string = proto_class.render_py().strip()
        expected = """@dataclasses.dataclass
class AllTypes(plasm.DataclassBase):
    __proto_cls__ = pb2.AllTypes
    my_string: str = dataclasses.field(default='', metadata={'dictator': dictators.BaseDictator})
    my_float: float = dataclasses.field(default=0.0, metadata={'dictator': dictators.BaseDictator})
    my_double: float = dataclasses.field(default=0.0, metadata={'dictator': dictators.BaseDictator})
    my_int32: int = dataclasses.field(default=0, metadata={'dictator': dictators.BaseDictator})
    my_int64: int = dataclasses.field(default=0, metadata={'dictator': dictators.LongDictator})
    my_uint32: int = dataclasses.field(default=0, metadata={'dictator': dictators.BaseDictator})
    my_uint64: int = dataclasses.field(default=0, metadata={'dictator': dictators.LongDictator})
    my_sint32: int = dataclasses.field(default=0, metadata={'dictator': dictators.BaseDictator})
    my_sint64: int = dataclasses.field(default=0, metadata={'dictator': dictators.LongDictator})
    my_fixed32: int = dataclasses.field(default=0, metadata={'dictator': dictators.BaseDictator})
    my_fixed64: int = dataclasses.field(default=0, metadata={'dictator': dictators.LongDictator})
    my_sfixed32: int = dataclasses.field(default=0, metadata={'dictator': dictators.BaseDictator})
    my_sfixed64: int = dataclasses.field(default=0, metadata={'dictator': dictators.LongDictator})
    my_bool: bool = dataclasses.field(default=False, metadata={'dictator': dictators.BaseDictator})
    my_bytes: bytes = dataclasses.field(default=b'', metadata={'dictator': dictators.ByteDictator})"""
        self.assertEqual(expected, class_string)

    def test_all_types_list(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.AllTypesList.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        class_string = proto_class.render_py().strip()
        expected = """@dataclasses.dataclass
class AllTypesList(plasm.DataclassBase):
    __proto_cls__ = pb2.AllTypesList
    my_string_list: List[str] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_float_list: List[float] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_double_list: List[float] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_int32_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_int64_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.LongDictator, 'is_list': True})
    my_uint32_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_uint64_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.LongDictator, 'is_list': True})
    my_sint32_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_sint64_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.LongDictator, 'is_list': True})
    my_fixed32_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_fixed64_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.LongDictator, 'is_list': True})
    my_sfixed32_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_sfixed64_list: List[int] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.LongDictator, 'is_list': True})
    my_bool_list: List[bool] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.BaseDictator, 'is_list': True})
    my_bytes_list: List[bytes] = dataclasses.field(default_factory=list, metadata={'dictator': dictators.ByteDictator, 'is_list': True})"""
        self.assertEqual(expected, class_string)

    def test_all_types_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.AllTypesMap.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        class_string = proto_class.render_py().strip()
        expected = """@dataclasses.dataclass
class AllTypesMap(plasm.DataclassBase):
    __proto_cls__ = pb2.AllTypesMap
    my_string_map: Dict[str, str] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_int32_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_int64_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.LongDictator, 'is_map': True})
    my_uint32_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_uint64_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.LongDictator, 'is_map': True})
    my_sint32_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_sint64_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.LongDictator, 'is_map': True})
    my_fixed32_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_fixed64_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.LongDictator, 'is_map': True})
    my_sfixed32_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})
    my_sfixed64_map: Dict[int, int] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.LongDictator, 'is_map': True})
    my_bool_map: Dict[bool, bool] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True})"""
        self.assertEqual(expected, class_string)

    def test_all_types_nested_map(self):
        from sandbox.test import rainbow_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(rainbow_pb2.AllTypesNestedMap.DESCRIPTOR, neobuilder.generators.symbols.modules.ProtoModule(rainbow_pb2))
        class_string = proto_class.render_py().strip()
        expected = """@dataclasses.dataclass
class AllTypesNestedMap(plasm.DataclassBase):
    __proto_cls__ = pb2.AllTypesNestedMap
    my_string_map: Dict[str, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_int32_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_int64_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_uint32_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_uint64_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_sint32_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_sint64_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_fixed32_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_fixed64_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_sfixed32_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_sfixed64_map: Dict[int, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})
    my_bool_map: Dict[bool, SubMessage] = dataclasses.field(default_factory=dict, metadata={'dictator': dictators.BaseDictator, 'is_map': True, 'is_obj': True})"""
        self.assertEqual(expected, class_string)

    def test_duration(self):
        from sandbox.test import timeduration_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(timeduration_pb2.DurationMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(timeduration_pb2))
        exp_field = 'my_duration'
        exp_type_hint = 'datetime.timedelta'
        exp_dictator = 'DurationDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}"
        exp_default = 'default=None'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_duration_list(self):
        from sandbox.test import timeduration_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(timeduration_pb2.DurationMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(timeduration_pb2))
        exp_field = 'my_duration_list'
        exp_type_hint = 'List[datetime.timedelta]'
        exp_dictator = 'DurationDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True"
        exp_default = 'default_factory=list'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})",
                         proto_field.render_py())

    def test_duration_map(self):
        from sandbox.test import timeduration_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(timeduration_pb2.DurationMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(timeduration_pb2))
        exp_field = 'my_duration_map'
        exp_type_hint = 'Dict[str, datetime.timedelta]'
        exp_dictator = 'DurationDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True"
        exp_default = 'default_factory=dict'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})",
                         proto_field.render_py())


class ProtoAnyFieldTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        sys.path.append(BUILD_ROOT)

    def test_any(self):
        from sandbox.test import anytest_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(anytest_pb2.AnyMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(anytest_pb2))
        exp_field = 'my_any'
        exp_type_hint = 'plasm.DataclassBase'
        exp_dictator = 'AnyDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}"
        exp_default = 'default=None'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_any())
        self.assertFalse(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_any_list(self):
        from sandbox.test import anytest_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(anytest_pb2.AnyMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(anytest_pb2))
        exp_field = 'my_any_list'
        exp_type_hint = 'List[plasm.DataclassBase]'
        exp_dictator = 'AnyDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True"
        exp_default = 'default_factory=list'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_any())
        self.assertFalse(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})",
                         proto_field.render_py())

    def test_any_map(self):
        from sandbox.test import anytest_pb2
        proto_class = neobuilder.generators.symbols.dataclass.ProtoClass(anytest_pb2.AnyMessage.DESCRIPTOR,
                                                                         neobuilder.generators.symbols.modules.ProtoModule(anytest_pb2))
        exp_field = 'my_any_map'
        exp_type_hint = 'Dict[str, plasm.DataclassBase]'
        exp_dictator = 'AnyDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True"
        exp_default = 'default_factory=dict'

        proto_field = neobuilder.generators.symbols.dataclass.ProtoField(proto_class.descriptor.fields_by_name[exp_field], proto_class)

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertTrue(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())

        self.assertTrue(proto_field.is_any())
        self.assertFalse(proto_field.is_duration())
        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())
        self.assertFalse(proto_field.is_enum())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})",
                         proto_field.render_py())


class ProtoEnumFieldTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        sys.path.append(BUILD_ROOT)

    def test_external_enum(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_enum = proto_module.enum_map['ExternalEnum']
        self.assertEqual('ExternalEnum', proto_enum.get_name())
        self.assertEqual('ExternalEnum', proto_enum.get_class_import_name())
        self.assertEqual("""THREE = ExternalEnum(3)
TWO = ExternalEnum(2)
ONE = ExternalEnum(1)
ZERO_AND_DEFAULT = ExternalEnum(0)""", proto_enum.render_py_consts().strip())
        self.assertEqual("""class ExternalEnum(enum.IntEnum):
    __proto_cls__ = pb2.ExternalEnum
    THREE = 3
    TWO = 2
    ONE = 1
    ZERO_AND_DEFAULT = 0""", proto_enum.render_py_class().strip())

    def test_external_enum_alias(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_enum = proto_module.enum_map['ExternalAliasEnum']
        self.assertEqual('ExternalAliasEnum', proto_enum.get_name())
        self.assertEqual('ExternalAliasEnum', proto_enum.get_class_import_name())
        self.assertEqual("""SEX = ExternalAliasEnum(3)
SIX = ExternalAliasEnum(3)
FIMM = ExternalAliasEnum(2)
FIVE = ExternalAliasEnum(2)
FJORIR = ExternalAliasEnum(1)
FOUR = ExternalAliasEnum(1)
ZERO = ExternalAliasEnum(0)
DEFAULT = ExternalAliasEnum(0)""", proto_enum.render_py_consts().strip())
        self.assertEqual("""class ExternalAliasEnum(enum.IntEnum):
    __proto_cls__ = pb2.ExternalAliasEnum
    SEX = 3
    SIX = 3
    FIMM = 2
    FIVE = 2
    FJORIR = 1
    FOUR = 1
    ZERO = 0
    DEFAULT = 0""", proto_enum.render_py_class().strip())

    def test_external_enum_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        proto_field = proto_class.field_map['my_enum']

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual('my_enum', proto_field.py_name())
        self.assertEqual('ExternalEnum', proto_field.get_type_hint())
        self.assertEqual('EnumDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.EnumDictator, 'is_enum': True", proto_field.get_metadata())
        self.assertEqual("my_enum: ExternalEnum = dataclasses.field(default=0, metadata={'dictator': dictators.EnumDictator, 'is_enum': True})", proto_field.render_py())

    def test_external_enum_alias_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        proto_field = proto_class.field_map['my_alias_enum']

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual('my_alias_enum', proto_field.py_name())
        self.assertEqual('ExternalAliasEnum', proto_field.get_type_hint())
        self.assertEqual('EnumDictator', proto_field.get_dictator_name())
        self.assertEqual("'dictator': dictators.EnumDictator, 'is_enum': True", proto_field.get_metadata())
        self.assertEqual("my_alias_enum: ExternalAliasEnum = dataclasses.field(default=0, metadata={'dictator': dictators.EnumDictator, 'is_enum': True})", proto_field.render_py())

    def test_external_enum_list_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        exp_field = 'my_enum_list'
        exp_dictator = 'EnumDictator'
        exp_type_hint = 'List[ExternalEnum]'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True, 'is_enum': True"
        exp_default = 'default_factory=list'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_external_enum_alias_list_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        exp_field = 'my_alias_enum_list'
        exp_dictator = 'EnumDictator'
        exp_type_hint = 'List[ExternalAliasEnum]'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True, 'is_enum': True"
        exp_default = 'default_factory=list'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_external_enum_map_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        exp_field = 'my_enum_map'
        exp_dictator = 'EnumDictator'
        exp_type_hint = 'Dict[str, ExternalEnum]'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True, 'is_enum': True"
        exp_default = 'default_factory=dict'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_external_enum_alias_map_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithExternalEnum']
        exp_field = 'my_alias_enum_map'
        exp_dictator = 'EnumDictator'
        exp_type_hint = 'Dict[str, ExternalAliasEnum]'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True, 'is_enum': True"
        exp_default = 'default_factory=dict'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']
        proto_enum = proto_class.enum_map['InternalEnum']
        self.assertEqual('InternalEnum', proto_enum.get_name())
        self.assertEqual('WithInternalEnum.InternalEnum', proto_enum.get_class_import_name())
        self.assertEqual("""SIX = InternalEnum(6)
FIVE = InternalEnum(5)
FOUR = InternalEnum(4)
ZERO_AND_DEFAULT = InternalEnum(0)""", proto_enum.render_py_consts().strip())
        self.assertEqual("""class InternalEnum(enum.IntEnum):
    __proto_cls__ = pb2.WithInternalEnum.InternalEnum
    SIX = 6
    FIVE = 5
    FOUR = 4
    ZERO_AND_DEFAULT = 0""", proto_enum.render_py_class().strip())

        self.assertEqual("""SIX = InternalEnum(6)
    FIVE = InternalEnum(5)
    FOUR = InternalEnum(4)
    ZERO_AND_DEFAULT = InternalEnum(0)""", proto_enum.render_py_consts(1).strip())
        self.assertEqual("""class InternalEnum(enum.IntEnum):
        __proto_cls__ = pb2.WithInternalEnum.InternalEnum
        SIX = 6
        FIVE = 5
        FOUR = 4
        ZERO_AND_DEFAULT = 0""", proto_enum.render_py_class(1).strip())

    def test_internal_enum_alias(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']
        proto_enum = proto_class.enum_map['InternalAliasEnum']
        self.assertEqual('InternalAliasEnum', proto_enum.get_name())
        self.assertEqual('WithInternalEnum.InternalAliasEnum', proto_enum.get_class_import_name())
        self.assertEqual("""NIU = InternalAliasEnum(9)
NINE = InternalAliasEnum(9)
ATTA = InternalAliasEnum(8)
EIGHT = InternalAliasEnum(8)
SJO = InternalAliasEnum(7)
SEVEN = InternalAliasEnum(7)
ZERO = InternalAliasEnum(0)
DEFAULT = InternalAliasEnum(0)""", proto_enum.render_py_consts().strip())
        self.assertEqual("""class InternalAliasEnum(enum.IntEnum):
    __proto_cls__ = pb2.WithInternalEnum.InternalAliasEnum
    NIU = 9
    NINE = 9
    ATTA = 8
    EIGHT = 8
    SJO = 7
    SEVEN = 7
    ZERO = 0
    DEFAULT = 0""", proto_enum.render_py_class().strip())

        self.assertEqual("""NIU = InternalAliasEnum(9)
    NINE = InternalAliasEnum(9)
    ATTA = InternalAliasEnum(8)
    EIGHT = InternalAliasEnum(8)
    SJO = InternalAliasEnum(7)
    SEVEN = InternalAliasEnum(7)
    ZERO = InternalAliasEnum(0)
    DEFAULT = InternalAliasEnum(0)""", proto_enum.render_py_consts(1).strip())
        self.assertEqual("""class InternalAliasEnum(enum.IntEnum):
        __proto_cls__ = pb2.WithInternalEnum.InternalAliasEnum
        NIU = 9
        NINE = 9
        ATTA = 8
        EIGHT = 8
        SJO = 7
        SEVEN = 7
        ZERO = 0
        DEFAULT = 0""", proto_enum.render_py_class(1).strip())

    def test_internal_enum_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_enum'
        exp_dictator = 'EnumDictator'
        exp_type_hint = 'WithInternalEnum.InternalEnum'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_enum': True"
        exp_default = 'default=0'

        proto_field = proto_class.field_map[exp_field]

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum_alias_field(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_alias_enum'
        exp_type_hint = 'WithInternalEnum.InternalAliasEnum'
        exp_dictator = 'EnumDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_enum': True"
        exp_default = 'default=0'

        proto_field = proto_class.field_map[exp_field]

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum_list(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_enum_list'
        exp_type_hint = 'List[WithInternalEnum.InternalEnum]'
        exp_dictator = 'EnumDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True, 'is_enum': True"
        exp_default = 'default_factory=list'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum_alias_list(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_alias_enum_list'
        exp_type_hint = 'List[WithInternalEnum.InternalAliasEnum]'
        exp_dictator = 'EnumDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_list': True, 'is_enum': True"
        exp_default = 'default_factory=list'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertTrue(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum_map(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_enum_map'
        exp_type_hint = 'Dict[str, WithInternalEnum.InternalEnum]'
        exp_dictator = 'EnumDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True, 'is_enum': True"
        exp_default = 'default_factory=dict'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_internal_enum_alias_map(self):
        from sandbox.test import enums_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(enums_pb2)
        proto_class = proto_module.message_map['WithInternalEnum']

        exp_field = 'my_internal_alias_enum_map'
        exp_type_hint = 'Dict[str, WithInternalEnum.InternalAliasEnum]'
        exp_dictator = 'EnumDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_map': True, 'is_enum': True"
        exp_default = 'default_factory=dict'

        proto_field = proto_class.field_map[exp_field]

        self.assertFalse(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertTrue(proto_field.is_map())

        self.assertTrue(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertFalse(proto_field.is_message())
        self.assertTrue(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

    def test_foreign_import_type(self):
        from sandbox.test import beta_pb2
        proto_module = neobuilder.generators.symbols.modules.ProtoModule(beta_pb2)
        proto_class = proto_module.message_map['BetaMessage']

        exp_field = 'my_foreign_message'
        exp_type_hint = 'sandbox__test__alpha_dc.AlphaMessage'
        exp_dictator = 'BaseDictator'
        exp_meta = f"'dictator': dictators.{exp_dictator}, 'is_obj': True"
        exp_default = 'default=None'

        proto_field = proto_class.field_map[exp_field]

        self.assertTrue(proto_field.is_single())
        self.assertFalse(proto_field.is_list())
        self.assertFalse(proto_field.is_map())

        self.assertFalse(proto_field.is_proto_repeated())

        self.assertFalse(proto_field.is_scalar())
        self.assertTrue(proto_field.is_message())
        self.assertFalse(proto_field.is_enum())

        self.assertFalse(proto_field.is_timestamp())
        self.assertFalse(proto_field.is_bytes())

        self.assertEqual(exp_field, proto_field.py_name())
        self.assertEqual(exp_type_hint, proto_field.get_type_hint())
        self.assertEqual(exp_dictator, proto_field.get_dictator_name())
        self.assertEqual(exp_meta, proto_field.get_metadata())
        self.assertEqual(f"{exp_field}: {exp_type_hint} = dataclasses.field({exp_default}, metadata={{{exp_meta}}})", proto_field.render_py())

        self.assertEqual('from sandbox.test import alpha_dc as sandbox__test__alpha_dc\n', proto_module.render_imports())



from __future__ import annotations
__all__ = [
    'ProtoModule',
]

import datetime
import hashlib
import os
import re
from typing import *

from google.protobuf import json_format

from neobuilder.generators.symbols import service
from neobuilder.generators.symbols import dataclass
from neobuilder.generators import servicebuilders
from neobuilder.neobuilder.render import TemplateRenderer

import logging
log = logging.getLogger(__name__)


CHECKSUM_MATCHER = re.compile(r'^# md5: (?P<checksum>[\dA-F]{32})', re.IGNORECASE)


class ProtoModule(object):
    def __init__(self, module, indent='    '):
        self.module = module
        self.indent = indent
        self.descriptor: json_format.descriptor.FileDescriptor = module.DESCRIPTOR
        self.message_map: Mapping[str, dataclass.ProtoClass] = {}
        self.enum_map: Mapping[str, dataclass.ProtoEnum] = {}
        self.service_map: Mapping[str, service.ProtoService] = {}
        self.imports: Mapping[str, json_format.descriptor.FileDescriptor] = {}

        self._load_module()

    def _load_module(self):
        self.enum_map = {n: dataclass.ProtoEnum(des, self) for n, des in self.descriptor.enum_types_by_name.items()}
        self.message_map = {n: dataclass.ProtoClass(des, self) for n, des in self.descriptor.message_types_by_name.items()}
        self.service_map = {n: service.ProtoService(des, self) for n, des in self.descriptor.services_by_name.items()}

    def get_module_full_name(self):
        return self.module.__name__

    def get_module_name(self):
        return self.module.__name__.split('.')[-1]

    def get_module_name_parts(self):
        return self.module.__name__.split('.')

    def get_package(self):
        return self.module.__package__

    def get_file_full_name(self):
        return self.module.__file__

    def get_file_name(self):
        return os.path.basename(self.get_file_full_name())

    def get_file_path(self):
        return os.path.dirname(self.get_file_full_name())

    def get_name(self):
        return os.path.splitext(self.get_file_name())[0]

    def get_render_file_name(self):
        return f'{self.get_file_full_name()[:-7]}_dc.py'

    def get_grpc_file_name(self):
        return f'{self.get_file_full_name()[:-3]}_grpc.py'

    def get_md5_hash(self, contents):
        hasher = hashlib.md5()
        if isinstance(contents, str):
            contents = contents.encode('utf-8')
        hasher.update(contents)
        return hasher.hexdigest()

    def can_write(self, filename, silent=False):
        if not os.path.exists(filename):
            return True  # File doesn't exist!

        content = []
        with open(filename, 'r', encoding='utf-8') as fin:
            content = [l for l in fin]

        if not content:
            return True  # File is empty

        first = content[0]
        m = CHECKSUM_MATCHER.match(first)
        if not m:
            if not silent:
                log.warning('Checksum not found for file: %s - It may have been manually edited.', filename)
            return False  # Checksum not found!

        found_sum = m.group('checksum')
        file_sum = self.get_md5_hash(''.join(content[1:]))

        if not found_sum == file_sum:
            if not silent:
                log.warning('Checksum mismatch in file: %s - It may have been manually edited.', filename)
            return False  # Checksum mismatch!

        return True  # Otherwise we're good :)

    def is_grpc_file(self):
        if os.path.exists(self.get_grpc_file_name()):
            return True
        return False

    def write_rendered_file(self):
        self.write_dataclass_file()
        if self.service_map:
            self.write_api_file()
            if self.is_grpc_file():
                self.write_grpc_file()
                self.write_grpc_impl_file()
                # TODO(thordurm@ccpgames.com>): This is unnecessary
                # self.write_grpc_server_file()
            # TODO(thordurm@ccpgames.com>): This is unnecessary
            # self.write_impl_file()

    def write_file(self, file_name, content, silent=False):
        # TODO(thordurm@ccpgames.com>): Need to support generating multiple dicectories if needed
        # TODO(thordurm@ccpgames.com>): Need to ensure file paths make sense if proto modules are libraries
        checksum = self.get_md5_hash(content)
        if self.can_write(file_name, silent):
            with open(file_name, 'w', encoding='utf-8') as fout:
                fout.write(f'# md5: {checksum}\n')
                fout.write(content)

    #
    # Dataclass stuff!
    #

    def render_proto_import(self):
        return f'from {self.get_package()} import {self.get_module_name()} as pb2'

    def render_imports(self):
        import_set = set()
        for m in self.message_map.values():
            import_set.update(m.get_imports())
        if import_set:
            return '\n'.join(sorted(import_set)) + '\n'
        return ''

    def get_all_list(self) -> List[str]:
        all_list = [en.get_name() for en in self.enum_map.values()]
        all_list.extend([pc.get_class_name() for pc in self.message_map.values()])
        return all_list

    def render_dataclass_file(self):
        return TemplateRenderer().render_file('dataclass_module', **{
            'module_full_name': self.get_module_full_name(),
            'iso_timestamp': datetime.datetime.now().isoformat(),
            'imports': self.render_imports(),
            'proto_import': self.render_proto_import(),
            'enum_list': [en.render_py() for en in self.enum_map.values()],
            'message_list': [pc.render_py() for pc in self.message_map.values()],
            'all_list': self.get_all_list(),
        })

    def write_dataclass_file(self):
        log.debug(f' - - write_dataclass_file() {self.get_render_file_name()}')
        self.write_file(self.get_render_file_name(), self.render_dataclass_file())

    #
    # GRPC SERVER FILE!
    #

    def write_grpc_server_file(self):
        b = servicebuilders.get_module_builder('grpc_server')(self)
        self.write_file(b.get_render_filename(), b.render(), silent=True)

    def render_grpc_server_file(self):
        return self.render('grpc_server')

    #
    # API FILE!
    #

    def write_api_file(self):
        b = servicebuilders.get_module_builder('interface')(self)
        log.debug(f' - - write_api_file() {b.get_render_filename()}')
        self.write_file(b.get_render_filename(), b.render())

    def render_api_file(self):
        return self.render('interface')

    #
    # GRPC FILE! - grpc_receiver
    #

    def write_grpc_file(self):
        b = servicebuilders.get_module_builder('grpc_receiver')(self)
        log.debug(f' - - write_grpc_file() {b.get_render_filename()}')
        self.write_file(b.get_render_filename(), b.render())

    def render_grpc_file(self):
        return self.render('grpc_receiver')

    #
    # GRPC IMPL FILE!
    #

    def write_grpc_impl_file(self):
        b = servicebuilders.get_module_builder('grpc_sender')(self)
        log.debug(f' - - write_grpc_impl_file() {b.get_render_filename()}')
        self.write_file(b.get_render_filename(), b.render())

    def render_grpc_impl_file(self):
        return self.render('grpc_sender')

    #
    # Implementation FILE!
    #

    def write_impl_file(self):
        b = servicebuilders.get_module_builder('implementation')(self)
        self.write_file(b.get_render_filename(), b.render(), silent=True)

    def render_impl_file(self):
        return self.render('implementation')

    #
    # Generic Render
    #

    def render(self, builder_name: str):
        return servicebuilders.get_module_builder(builder_name)(self).render()
